"""
This module implements a general Logger aspect for logging method calls.

@author Ernesto Posse
"""

from abc import ABCMeta
import logging.config

from libutilitaspy import get_config_file
import core


DEFAULT_LOGGING_CONFIG_FILE = get_config_file()


class Logger(core.Aspect):

    classes = '.*'
    methods = '(?!__init__)(?!__str__)(?!__repr__)'     # Must leave these out, otherwise, an infinite recursion occurs
    
    def __init__(self, logger_name='', logging_config_file=DEFAULT_LOGGING_CONFIG_FILE, disable_existing_loggers=False):
        logging.config.fileConfig(logging_config_file, disable_existing_loggers=disable_existing_loggers)
        self.logger = logging.getLogger(logger_name)
        self.level = logging.DEBUG
        
    def before(self, klass, method, obj, *args, **kwargs):
        args_str_list = [str(arg) for arg in args]
        kwargs_str_list = ['%s = %s' % (key, arg) for key, arg in kwargs.items()]
        flattened_args = ', '.join([str(obj)] + args_str_list + kwargs_str_list)
        self.logger.log(self.level, "before %s.%s(%s)" % (klass.__name__, method.__name__, flattened_args))
    
    def after(self, klass, method, obj, retval, exc_type, exc_val, traceback):
        if exc_val is None:
            result = "returns %s" % (retval)
        else:
            result = "raises %s(%s)" % (exc_type.__name__, exc_val)
        self.logger.log(self.level, "after %s.%s(%s, ...) %s" % (klass.__name__, method.__name__, obj, result))


def logger_generator(klass, method, obj, logging_config_file, *args, **kwargs):
    logging.config.fileConfig(logging_config_file, disable_existing_loggers=False)
    logger = logging.getLogger('root')
    level = logging.DEBUG
    args_str_list = [str(arg) for arg in args]
    kwargs_str_list = ['%s = %s' % (key, arg) for key, arg in kwargs.items()]
    flattened_args = ', '.join([str(obj)] + args_str_list + kwargs_str_list)
    logger.log(level, "before %s.%s(%s)" % (klass.__name__, method.__name__, flattened_args))
    retval, exc_type, exc_val, _ = yield
    if exc_val is None:
        result = "returns %s" % (retval)
    else:
        result = "raises %s(%s)" % (exc_type.__name__, exc_val)
    logger.log(level, "after %s.%s(%s, ...) %s" % (klass.__name__, method.__name__, obj, result))

AltLogger = core.make_aspect_from_generator(logger_generator)

LoggerMetaClass = core.WeaverMetaClassFactory(Logger())

class LoggerAbstractMetaClass(LoggerMetaClass, ABCMeta): pass

