"""
Author: Ernesto Posse

Description:
    This module contains the categorical definitions of colimits. For a 
    formal definition see:
        http://en.wikipedia.org/wiki/Limit_(category_theory)
    
    Note: it contains the only the classes that represent colimits, not the
    algorithms that compute them. For these, see the 'diagrams' module.

    In particular, the argument to the CoLimit constructor is not the diagram
    of which the colimit is a colimit of. Instead the arguments are the elements
    that form a colimit, namely an object and a family of arrows, together with
    the function that computes the unique arrow to any other cone:
    
        CoLimit(object, arrows, unique_arrow_func, category)
        
    To obtain the colimit of a diagram, one must create an instance of diagram
    and invoke the colimit method on that instance. This method must return a
    CoLimit instance.

    New: for ease of use, I added a functional interface to compute colimits.
"""

from types import FunctionType, MethodType
from libutilitaspy.metaclassconfig import MetaClass
from libutilitaspy.data_structures.maps import Map
import diagrams


def colimit(diagram):
    return diagram.colimit()

def coproduct(pair):
    assert isinstance(pair, diagrams.Pair)
    return pair.colimit()

def coequalizer(parallel_arrows):
    assert isinstance(parallel_arrows, diagrams.ParallelArrows)
    return parallel_arrows.colimit()

def pushout(span):
    assert isinstance(span, diagrams.Span)
    return span.colimit()


class CoCone(object):
    
    __metaclass__ = MetaClass

    def __init__(self, obj, arrows):
        assert all(obj == phi.target for phi in arrows)
        self.category = obj.category
        self.vertex = obj
        self.arrows = arrows
        self.arrows_by_source = Map((phi.source, phi) for phi in arrows)

    def valid(self, diagram):
        assert isinstance(diagram, diagrams.Diagram)
        ##            Alternative:
        #        return all(phi.source in diagram.objects for phi in self.arrows) \
        #            and all(diagrams.Triangle(f, self.arrows_by_source[f.target], phi).commutes() \
        #                    for phi in self.arrows \
        #                    for f in phi.source.outgoing \
        #                    if f in diagram.arrows)
        return all(phi.source in diagram.objects for phi in self.arrows) \
            and all(diagrams.Triangle(f, phi, self.arrows_by_source[f.source]).commutes() \
                    for phi in self.arrows \
                    for f in phi.source.incoming \
                    if f in diagram.arrows)


class CoLimit(CoCone):
    
    __metaclass__ = MetaClass

    def __init__(self, obj, arrows, unique_arrow_func):
        """
        Parameters:
            obj: Object
            arrows: Arrow list
            unique_arrow_func: Cone -> Arrow
        Preconditions:
            1) all(obj == phi.source for phi in arrows)
                # obj is the source of all arrows
            2) all(phi.target in diagram.objects for phi in arrows)
                # the target of all arrows are in the diagram
            3) Cone(obj, arrows, diagram).valid()
                # (obj, arrows) form a valid Cone
            4) 
                # unique_arrow_func is a function that given any cone (X,psi),
                # returns the unique arrow from X to obj that makes the
                # resulting diagrams commute
        """
        CoCone.__init__(self, obj, arrows)
        assert type(unique_arrow_func) in (FunctionType, MethodType)
        self.unique_arrow_func = unique_arrow_func
    
    def unique_arrow(self, cocone):
        assert isinstance(cocone, CoCone)
        u = self.unique_arrow_func(cocone)
        assert u.source == self.vertex
        assert u.target == cocone.vertex
#        assert all(diagrams.Triangle(self.arrows_by_source[psi.source], u, psi).commutes() for psi in cocone.arrows)
        return u


class Initial(CoLimit):
    
    __metaclass__ = MetaClass
    
    def __init__(self, obj, unique_arrow_func):
        CoLimit.__init__(self, obj, [], unique_arrow_func)
    

class BinCoProduct(CoLimit):
    
    __metaclass__ = MetaClass

    def __init__(self, P, p1, p2, unique_arrow_func):
        CoLimit.__init__(self, P, [p1, p2], unique_arrow_func)
        self.inj1 = p1
        self.inj2 = p2

    
class CoEqualizer(CoLimit):
             
    __metaclass__ = MetaClass
    
    def __init__(self, Q, q, unique_arrow_func):
        CoLimit.__init__(self, Q, [q], unique_arrow_func)
        self.coeq = q
        

class Pushout(CoLimit):
    
    __metaclass__ = MetaClass
    
    def __init__(self, P, p1, p2, unique_arrow_func):
        CoLimit.__init__(self, P, [p1, p2], unique_arrow_func)
