"""
Author: Ernesto Posse

Description:
    This module contains the categorical definitions of diagrams. For a 
    formal definition see:
        http://en.wikipedia.org/wiki/Diagram_(category_theory)
    
    Note: it includes the algorithms that compute pullbacks in terms of 
    products and equalizers, and pushouts in terms of coproducts and coequalizers.
    Hence, a specific category does not need to implement either a pullback nor
    a pushout method; it only needs to implement product, coproduct, equalizer and
    coequalizer. In order to do that, the client must implement subclasses or
    Pair with methods product and coproduct, and ParallelArrows with methods
    equalizer and coequalizer
"""

from libutilitaspy.metaclassconfig import MetaClass
from libutilitaspy.data_structures.maps import Map
from libutilitaspy.data_structures.graphs import Node, Edge, Graph, GraphHomomorphism
import categories
import limits
import colimits


class Triangle(object):
    
    __metaclass__ = MetaClass

    def __init__(self, f, g, h):
        assert isinstance(f, categories.Arrow)
        assert isinstance(g, categories.Arrow)
        assert isinstance(h, categories.Arrow)
        assert f.target == g.source
        assert h.source == f.source
        assert h.target == g.target
        self.f = f
        self.g = g
        self.h = h
    
    def commutes(self):
        """
        Checks whether the triangle given by (f,g,h) commutes
        Returns:
            True iff h = g * f
        """
        return self.h == self.g * self.f


class Diagram(GraphHomomorphism):
    
    __metaclass__ = MetaClass

    """
    n0 = Node(0)
    n1 = Node(1)
    n2 = Node(2)
    e1 = Edge(n0, n1)
    e2 = Edge(n0, n1)
    pairshape = Graph([n0,n1],[])
    parallelshape = Graph([n0, n1], [e1, e2])
    pair_to = Graph([n0,n1,n2],[Edge(n0,n2),Edge(n1,n2)])
    D1 = Diagram(pairshape, Set, Map({n0: set0, n1: set1}, Map({}))
    """

    def __init__(self, shape, category, objectmap, arrowmap):
        assert isinstance(shape, Graph)
        assert isinstance(category, categories.FiniteCategory)
        assert isinstance(objectmap, Map)
        assert isinstance(arrowmap, Map)
        self.shape = shape
        self.category = category
        self.objectmap = objectmap
        self.arrowmap = arrowmap
        GraphHomomorphism.__init__(self, shape, category, objectmap, arrowmap)
    
    def commutes(self):
        """Checks if the whole diagram commutes"""
#        for f in self.arrows:
#            for g in f.target.outgoing:
#                if g in self.arrows:
#                    for h in self.category.hom(f.source, g.target):
#                        if h != g * f:
#                            return False
#        return True
        triangle_groups = (self.triangles(f,g) for f in self.arrows for g in self.arrows)
        triangles = reduce((lambda x, y: x | y), triangle_groups, set())
        return all(tri.commutes() for tri in triangles)

    def triangles(self, f, g):
        assert f in self.arrows
        assert g in self.arrows
        if f.target == g.source:
            return set(Triangle(f, g, h) for h in self.category.hom(f.source, g.target))
#            return set(Triangle(f,g,h) for h in self.arrows if h.source == f.source and h.target == g.target)
        else:
            return set()
    
#    def limit(self):
#        """Abstract method. Should return the limit of the diagram."""
#        pass
#    
#    def colimit(self):
#        """Abstract method. Should return the colimit of the diagram."""
#        pass
    
#    def is_object_pair(self):
#        return len(self.objects) == 2 and len(self.arrows) == 0
#    
#    def is_parallel_arrows(self):
#        return len(self.arrows) == 2 \
#            and self.arrows[0].source == self.arrows[1].source \
#            and self.arrows[0].target == self.arrows[1].target 
#    
#    def is_cone(self):
#        return len(self.arrows) == 2 \
#            and self.arrows[0].source == self.arrows[1].source 
#        
#    def is_cocone(self):
#        return len(self.arrows) == 2 \
#            and self.arrows[0].target == self.arrows[1].target 
        
    
class Empty(Diagram):
    
    __metaclass__ = MetaClass

    SHAPE = Graph([], [])
    
    def __init__(self, category):
        assert isinstance(category, categories.Category)
        objectmap = Map({})
        arrowmap = Map({})
        Diagram.__init__(self, Empty.SHAPE, category, objectmap, arrowmap)
    
    def limit(self):
        T = self.final()
        assert isinstance(T, limits.Final)
        return T

    def colimit(self):
        T = self.initial()
        assert isinstance(T, colimits.Initial)
        return T
    
#    def initial(self):
#        """Abstract method: should return an initial object in the category."""
#        pass
#
#    def final(self):
#        """Abstract method: should return an final object in the category."""
#        pass


class Pair(Diagram):

    __metaclass__ = MetaClass

    NODE0 = Node(0)
    NODE1 = Node(1)

    SHAPE = Graph([NODE0, NODE1], [])
    
    def __init__(self, A, B, category):
        assert isinstance(category, categories.Category)
        assert category.valid_object(A)
        assert category.valid_object(B)
        objectmap = Map({Pair.NODE0: A, Pair.NODE1: B})
        arrowmap = Map({})
        Diagram.__init__(self, Pair.SHAPE, category, objectmap, arrowmap)

    def limit(self):
        P = self.product()
        assert isinstance(P, limits.BinProduct)
        return P
    
    def colimit(self):
        P = self.coproduct()
        assert isinstance(P, colimits.BinCoProduct)
        return P
    
#    def product(self):
#        """Abstract method: should return the product of A and B."""
#        pass
#    
#    def coproduct(self):
#        """Abstract method: should return the coproduct of A and B."""
#        pass
    

class ParallelArrows(Diagram):
    
    __metaclass__ = MetaClass

    NODE0 = Node(0)
    NODE1 = Node(1)

    EDGE0 = Edge(NODE0, NODE1)
    EDGE1 = Edge(NODE0, NODE1)

    SHAPE = Graph([NODE0, NODE1], [EDGE0, EDGE1])

    def __init__(self, f, g, category):
        assert isinstance(category, categories.Category)
        assert category.valid_arrow(f)
        assert category.valid_arrow(g)
        assert f.source == g.source
        assert f.target == g.target
        A, B = f.source, f.target
        self.source = A
        self.target = B
        objectmap = Map({ParallelArrows.NODE0: A, ParallelArrows.NODE1: B})
        arrowmap = Map({ParallelArrows.EDGE0: f, ParallelArrows.EDGE1: g})
        Diagram.__init__(self, ParallelArrows.SHAPE, category, objectmap, arrowmap)

    def limit(self):
        Q = self.equalizer()
        assert isinstance(Q, limits.Equalizer)
        return Q
    
    def colimit(self):
        Q = self.coequalizer()
        assert isinstance(Q, colimits.CoEqualizer)
        return Q
    
#    def equalizer(self):
#        """Abstract method: should return the equalizer of A and B."""
#        pass
#    
#    def coequalizer(self):
#        """Abstract method: should return the coequalizer of A and B."""
#        pass
        

class CoSpan(Diagram):
    
    __metaclass__ = MetaClass

    NODE0 = Node(0)
    NODE1 = Node(1)
    NODE2 = Node(2)
    
    EDGE0 = Edge(NODE0, NODE2)
    EDGE1 = Edge(NODE1, NODE2)
    
    SHAPE = Graph([NODE0, NODE1, NODE2], [EDGE0, EDGE1])
    
    def __init__(self, f, g, category):
        assert isinstance(category, categories.Category)
        assert category.valid_arrow(f)
        assert category.valid_arrow(g)
        assert f.target == g.target
        A, B = f.source, g.source
        C = f.target
        objectmap = Map({CoSpan.NODE0: A, CoSpan.NODE1: B, CoSpan.NODE2: C})
        arrowmap = Map({CoSpan.EDGE0: f, CoSpan.EDGE1: g})
        Diagram.__init__(self, CoSpan.SHAPE, category, objectmap, arrowmap)

    def limit(self):
        Q = self.pullback()
        assert isinstance(Q, limits.Pullback)
        return Q
    
    def pullback(self):
        """Computes the pullback of A -f-> C <-g- B in the category."""
        A = self.objectmap[CoSpan.NODE0]
        B = self.objectmap[CoSpan.NODE1]
        f = self.arrowmap[CoSpan.EDGE0]
        g = self.arrowmap[CoSpan.EDGE1]
        P = self.category.product(A, B)
        assert isinstance(P, limits.BinProduct)
        Q = self.category.equalizer(f * P.proj1, g * P.proj2)
        assert isinstance(Q, limits.Equalizer)
        q1 = P.proj1 * Q.eq
        q2 = P.proj2 * Q.eq
        def unique_arrow_func(cone):
            assert isinstance(cone, limits.Cone)
            assert len(cone.arrows) == 2
            return Q.unique_arrow(limits.Cone(cone.vertex, [P.unique_arrow(cone)], self.category))
        return limits.Pullback(Q, q1, q2, unique_arrow_func)
    

class Span(Diagram):
    
    __metaclass__ = MetaClass

    NODE0 = Node(0)
    NODE1 = Node(1)
    NODE2 = Node(2)
    
    EDGE0 = Edge(NODE2, NODE0)
    EDGE1 = Edge(NODE2, NODE1)
    
    SHAPE = Graph([NODE0, NODE1, NODE2], [EDGE0, EDGE1])
    
    def __init__(self, f, g, category):
        assert isinstance(category, categories.Category)
        assert category.valid_arrow(f)
        assert category.valid_arrow(g)
        assert f.source == g.source
        A, B = f.target, g.target
        C = f.source
        objectmap = Map({Span.NODE0: A, Span.NODE1: B, Span.NODE2: C})
        arrowmap = Map({Span.EDGE0: f, Span.EDGE1: g})
        Diagram.__init__(self, Span.SHAPE, category, objectmap, arrowmap)

    def colimit(self):
        Q = self.pushout()
        assert isinstance(Q, colimits.Pushout)
        return Q
    
    def pushout(self):
        """Computes the pushout of A <-f- C -g-> B in the category."""
        A = self.objectmap[Span.NODE0]
        B = self.objectmap[Span.NODE1]
        f = self.arrowmap[Span.EDGE0]
        g = self.arrowmap[Span.EDGE1]
        P = self.category.coproduct(A, B)
        assert isinstance(P, colimits.BinCoProduct)
        Q = self.coequalizer(P.inj1 * f, P.inj2 * g)
        assert isinstance(Q, colimits.CoEqualizer)
        q1 = Q.coeq * P.inj1
        q2 = Q.coeq * P.inj2
        def unique_arrow_func(cocone):
            assert isinstance(cocone, colimits.CoCone)
            assert len(cocone.arrows) == 2
            return Q.unique_arrow(colimits.CoCone(cocone.vertex, [P.unique_arrow(cocone)], self.category))
        return colimits.Pushout(Q, q1, q2, unique_arrow_func)

