"""
Author: Ernesto Posse
Description:
    Category of finite sets

Typical use:
    
    Cat = DefaultFiniteSetCategory
    S1 = Set(['a','b','c'])
    S2 = Set([0,1])
    f = Function(S1, S2, {'a':0, 'b':1, 'c':0})
    P = S1 * S2    # equivalent to    P = DefaultFiniteSetCategory.cartesian_product(S1,S2)
    D = S1 + S2    # equivalent to    P = DefaultFiniteSetCategory.disjoint_union(S1,S2)
    x = P.proj1(('b',0))
    y = S.inj1('c')
    z = S.inj2(1)
    u = P.unique_arrow(D.vertex)
    w = u(('a',1))
    g = Function(S1, S2, {'a':1, b:'1', 'c':0})
    Q1 = Cat.equalizer(f, g)
    Q2 = Cat.coequalizer(f, g)
    S3 = Set(['x','y','z'])
    h = Function(S3, S2, {'x':0, 'y':0, 'z':0})
    P2 = Cat.pullback(f, g)
    k = Function(S1, S3, {'a':'y', 'b':'y', 'c':'x'})
    P3 = Cat.pushout(f, k)

    Cat = FiniteSetCategory()
    Cat.add_object(S1)
    Cat.add_object(S2)
    Cat.add_arrow(f)
"""

from libutilitaspy.metaclassconfig import MetaClass
from libutilitaspy.data_structures.maps import Map
import categories
import diagrams
import limits
import colimits


def empty(category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.initial()

def singleton(val=None, category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.final(val)

def product(A, B, category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.product(A, B)
    
def coproduct(A, B, category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.coproduct(A, B)
    
def equalizer(f,g,category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.equalizer(f, g)

def coequalizer(f,g,category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.coequalizer(f, g)

def pullback(f,g,category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.pullback(f, g)

def pushout(f,g,category=None):
    if category is None:
        category = DefaultFiniteSetCategory
    return category.pushout(f, g)


class Set(categories.Object):
    
    __metaclass__ = MetaClass

    def __init__(self, iterable, category=None):
        frozen_set = frozenset(iterable)
        if category is None:
            category = DefaultFiniteSetCategory
        categories.Object.__init__(self, frozen_set, category)
    
    def __iter__(self):
        return iter(self.obj)
    
    def __len__(self):
        return len(self.obj)
    
    def __contains__(self, item):
        return item in self.obj
    
    def __le__(self, other):
        return self.obj <= other.obj
    
    def __lt__(self, other):
        return self.obj < other.obj
    
    def issubset(self, other):
        return self.obj.issubset(other.obj)
    
    def __ge__(self, other):
        return self.obj >= other.obj

    def __gt__(self, other):
        return self.obj > other.obj

    def issuperset(self, other):
        return self.obj.issuperset(other.obj)
    
    def union(self, other):
        return Set(self.obj.union(other.obj), self.category)
    
    def __or__(self, other):
        return Set(self.obj | other.obj, self.category)
    
    def intersection(self, other):
        return Set(self.obj.intersection(other.obj), self.category)

    def __and__(self, other):
        return Set(self.obj & other.obj, self.category)
    
    def difference(self, other):
        return Set(self.obj.difference(other.obj), self.category)

    def __sub__(self, other):
        return Set(self.obj - other.obj, self.category)

    def symmetric_difference(self, other):
        return Set(self.obj.symmetric_difference(other.obj), self.category)

    def __xor__(self, other):
        return Set(self.obj ^ other.obj, self.category)
    
    def __mul__(self, other):
        return self.category.cartesian_product(self, other)
    
    def __add__(self, other):
        return self.category.disjoint_union(self, other)

    def copy(self):
        return Set(self.obj.copy(), self.category)
    
    def __repr__(self):
        return repr(self.obj)
    
    def __str__(self):
        return str(self.obj)
    
    def get_any_element(self):
        return iter(self.obj).next()
    

class Function(categories.Arrow, Map):
    
    __metaclass__ = MetaClass

    def __init__(self, source, target, function, category=None):
        if category is None:
            category = DefaultFiniteSetCategory
        def func(x):
            assert x in source
            y = function(x)
            assert y in target
            return y
        categories.Arrow.__init__(self, source, target, func, category)
        Map.__init__(self, function)
    
    def __eq__(self, other): 
        return Map.__eq__(self, other)
    
#    def __mul__(self, other):
#        return Function(self.source, other.target, Map.__mul__(self, other), self.category)
#        
        
class SmallSetCategory(categories.Category):
    
    __metaclass__ = MetaClass

    def __init__(self):
        categories.Category.__init__(self, set, Function)
    
    def composition(self, f, g):
        # TODO: change function to explicit map composition
        return Function(f.source, g.target, Map.__mul__(f, g), self)
    
    def identity(self, s):
        return Function(s, s, lambda x: x, self)
    

class FiniteSetCategory(SmallSetCategory, categories.FiniteCategory):
    
    __metaclass__ = MetaClass

    def __init__(self, sets=None, functions=None):
        categories.FiniteCategory.__init__(self, Set, Function, sets, functions)

    def valid_object(self, obj):
        return isinstance(obj, Set) and categories.FiniteCategory.valid_object(self, obj)
    
    def valid_arrow(self, arrow):
        return isinstance(arrow, Function) and categories.FiniteCategory.valid_arrow(self, arrow)
    
    def initial(self):
        return Empty(self).initial()

    def final(self, val=0):
        return Empty(self).final(val)

    def product(self, A, B):
        return SetPair(A, B, self).limit()
    
    def coproduct(self, A, B):
        return SetPair(A, B, self).colimit()
    
    def equalizer(self, f, g):
        return ParallelFunctions(f, g, self).limit()
    
    def coequalizer(self, f, g):
        return ParallelFunctions(f, g, self).colimit()

    emptyset = initial
    singleton = final
    cartesian_product = product
    disjoint_union = coproduct


class Empty(diagrams.Empty):
    
    __metaclass__ = MetaClass
    
    def __init__(self, category):
        diagrams.Empty.__init__(self, category)

    def initial(self):
        empty = Set([], self.category)
        def unique_arrow_func(cocone):
            return Function(empty, cocone.vertex, Map({}), self.category)
        return colimits.Initial(empty, unique_arrow_func)

    def final(self, val=0):
        """The value contained in the singleton is optional."""
        singleton = Set([val], self.category)
        def unique_arrow_func(cone):
            return Function(cone.vertex, singleton, lambda x: val, self.category)
        return limits.Final(singleton, unique_arrow_func)
    
    
class SetPair(diagrams.Pair):

    __metaclass__ = MetaClass
    
    def __init__(self, A, B, category):
        assert isinstance(A, Set)
        assert isinstance(B, Set)
        diagrams.Pair.__init__(self, A, B, category)
    
    def product(self):
        A = self.objectmap[diagrams.Pair.NODE0]
        B = self.objectmap[diagrams.Pair.NODE1]
        AxB = Set(((a,b) for a in A for b in B), self.category)
        proj1 = Function(AxB, A, lambda pair: pair[0], self.category)
        proj2 = Function(AxB, B, lambda pair: pair[1], self.category)
        def unique_arrow_func(cone):
            def u(x):
                y = (cone.arrows_by_target[A](x), cone.arrows_by_target[B](x))
                assert y[0] in A and y[1] in B
                return y
#                return (cone.arrows[0](x), cone.arrows[1](x))
            return Function(cone.vertex, AxB, u, self.category)
        return limits.BinProduct(AxB, proj1, proj2, unique_arrow_func)
    
    def coproduct(self):
        A = self.objectmap[diagrams.Pair.NODE0]
        B = self.objectmap[diagrams.Pair.NODE1]
        AUB = Set(((0,a) for a in A), self.category) | Set(((1,b) for b in B), self.category)
        inj1 = Function(A, AUB, lambda a: (0,a), self.category)
        inj2 = Function(B, AUB, lambda b: (1,b), self.category)
        def unique_arrow_func(cocone):
            assert isinstance(cocone, colimits.CoCone)
            def u(x):
                if x[0] == 0:
                    y = cocone.arrows_by_source[A](x[1])
#                   return cocone.arrows[0](x[1])
                else:
                    y = cocone.arrows_by_source[B](x[1])
#                    return cocone.arrows[1](x[1])
                assert y in cocone.vertex
                return y
            return Function(AUB, cocone.vertex, u, self.category)
        return colimits.BinCoProduct(AUB, inj1, inj2, unique_arrow_func)


class ParallelFunctions(diagrams.ParallelArrows):

    __metaclass__ = MetaClass
    
    def __init__(self, f, g, category):
        assert isinstance(f, Function) 
        assert isinstance(g, Function)
        diagrams.ParallelArrows.__init__(self, f, g, category)

    def equalizer(self):
        A = self.objectmap[diagrams.ParallelArrows.NODE0]
        f = self.arrowmap[diagrams.ParallelArrows.EDGE0]
        g = self.arrowmap[diagrams.ParallelArrows.EDGE1]
        Q = Set((x for x in A if f(x) == g(x)), self.category)
        def restr(x):
            assert x in Q
            return x
        q = Function(Q, A, restr, self.category)
        def unique_arrow_func(cone):
            assert len(cone.arrows) == 1
            def u(x):
                y = cone.arrows_by_target[A](x)
                assert y in Q
                return y
            return Function(cone.vertex, Q, u, self.category)
        return limits.Equalizer(Q, q, unique_arrow_func)

    def coequalizer(self):
        A = self.objectmap[diagrams.ParallelArrows.NODE0]
        B = self.objectmap[diagrams.ParallelArrows.NODE1]
        f = self.arrowmap[diagrams.ParallelArrows.EDGE0]
        g = self.arrowmap[diagrams.ParallelArrows.EDGE1]
        assert isinstance(f, Function) and isinstance(g, Function)
        assert f.source == g.source
        assert f.target == g.target
        # compute partition tables
        part, ind, _, _ = self.__compute_coequalizer_tables(A, B, f, g) 
        Q = Set((Set(part[i], self.category) for i in part), self.category)
#        Q = Set((i for i in part), self.category)
        q = Function(B, Q, lambda b: Set(part[ind[b]], self.category), self.category)
#        q = Function(B, Q, lambda b: ind[b], self.category)
        def unique_arrow_func(cocone):
            assert len(cocone.arrows) == 1
            def u(x):
                # x is an equivalence class [b]_{~}
                assert len(x) > 0
                b = x.get_any_element()
#                b = x[0]
#                b = part[x][0]
                y = cocone.arrows_by_source[B](b)
                assert y in cocone.vertex
                return y
            return Function(Q, cocone.vertex, u, self.category)
        return colimits.CoEqualizer(Q, q, unique_arrow_func)
    
    def __compute_pre_coequalizer_tables(self, A, B, f, g):
        """
        Computes the pre-coequalizer tables for A ==f,g==> B where A and B are sets
        (or anything iterable) and f and g are functions from A to B.
            
        This is, it computes the pair (Q,q) where:
            a) Q is the quotient (partition) of set B with respect to an equivalence 
               relation implied by the binary relation R defined as follows:
            
                R = {(b1, b2) in BxB | there is an a in A such that b1 = f(a) and b2 = g(a)}
        
                The resulting quotient is
                
                    Q = B/R'
                
                where R' is the reflexive, symmetric and transitive closure of R. 
            b) q:B->Q is the cannonical function, mapping each elemebt b of B to
               its equivalence class (partition) [b] in Q.
                   
        This function computes the *pre* coequalizer rather than the full coequalizer 
        because it does not include those elements of B which are not the image of
        some element of A according to either f or g.
        
        Parameters:
            A: 'a set           Any iterable_object; typically a set or list.
            f: 'a -> 'b         Any function over A
            g: 'a -> 'b         Any function over B
            B: 'b set           Any iterable_object; typically a set or list.
        Returns:
            (partition_table, partition_index_table, preimage, last_index)
                : (int, 'b list) dict * ('b,int) dict * (int,'a list) dict * int
            where
                partition_table: (int, 'b list) dict
                    maps a number to a partition of B according to R (this 
                    corresponds to the quotient Q, see above)
                partition_index: ('b,int) dict
                    maps each element of B to the index of its partition (this 
                    corresponds to the cannonical map q, see above)
                preimage: (int,'a list) dict
                    maps each partition (index) to the elements of A which have
                    images (according to either f or g) in the partition.
                last_index: int
                    is the last index used in the partition table
        Preconditions:
            domain(f) = A
            domain(g) = A
        Postconditions:
            Assuming I is the set of indices in partition_table
            1) for all b in B, b in partition_table[partition_index[b]]
            2) B = union partition_table[i] for i in I
            3) for all i,j in I, if i is not j, the intersection of partition_table[i]
                and partition_table[j] is empty
            (this conditions essentially state that the table is in fact a partition of B)
        """
        partition_table = {}    # maps partition ids to partitions in B
        partition_index = {}    # maps elements of B to partition ids
        preimage = {}           # maps each partition index to the list of elements in A such that their image according to f or g is in the partition
        def add_preimage(j,a):
            if j in preimage:
                preimage[j].append(a)
            else:
                preimage[j] = [a]
        index = -1
        # invariant: for all b in B, b in partition_table[partition_index[b]]
        for a in A:
            b1 = f(a)
            b2 = g(a)
            if b1 in partition_index:
                i1 = partition_index[b1]
                b1_partition = partition_table[i1]
                if b2 in partition_index:
                    # both b1 and b2 are already in some partitions
                    # so we merge them if they are not the same partition.
                    # If they are in the same partition, then we don't need
                    # to do anything as they are already in place.
                    i2 = partition_index[b2]
                    b2_partition = partition_table[i2]
                    if i1 != i2:
                        b1_partition += b2_partition
                        for b in b2_partition:
                            partition_index[b] = i1
                        del partition_table[i2]
                        preimage[i1] += preimage[i2]
                        add_preimage(i1, a)
                        del preimage[i2]
                else:
                    # If b1 is in some partition, but b2 is not in any yet,
                    # we add b2 to b1's partition.
                    b1_partition.append(b2)
                    partition_index[b2] = i1
                    add_preimage(i1,a)
            else:
                if b2 in partition_index:
                    # If b2 is in some partition, but b1 is not in any yet,
                    # we add b1 to b2's partition.
                    i2 = partition_index[b2]
                    b2_partition = partition_table[i2]
                    b2_partition.append(b1)
                    partition_index[b1] = i2
                    add_preimage(i2,a)
                else:
                    # b1 and b2 are not in any partition yet
                    # so we create a new entry in the main table,
                    # and entries for b1 and b2 in the index table
                    index += 1
                    partition_index[b1] = index
                    partition_index[b2] = index
                    if b1 != b2:
                        partition_table[index] = [b1, b2]
                    else:
                        partition_table[index] = [b1]
                    add_preimage(index,a)
        return partition_table, partition_index, preimage, index

    def __compute_coequalizer_tables(self, A, B, f, g):
        """
        This is like __compute_pre_coequalizer_tables, but adds a partition for each
        element of B which is not the image of some element of A.
        """
        part, ind, preimg, index = self.__compute_pre_coequalizer_tables(A, B, f, g)
        # We add a partition for each element in B which is not already in
        # some partition, i.e. those elements of B which are not images of
        # some a in A either by f or by g.
        for b in B:
            if b not in ind:
                index += 1
                ind[b] = index
                part[index] = [b]
                preimg[index] = []
        return part, ind, preimg, index


DefaultFiniteSetCategory = FiniteSetCategory()

        

class SetCategoryException(Exception):
    pass
