"""
Author: Ernesto Posse

Description:
    This module contains the categorical definitions of limits. For a 
    formal definition see:
        http://en.wikipedia.org/wiki/Limit_(category_theory)
    
    Note: it contains the only the classes that represent limits, not the
    algorithms that compute them. For these, see the 'diagrams' module.
    
    In particular, the argument to the Limit constructor is not the diagram
    of which the limit is a limit of. Instead the arguments are the elements
    that form a limit, namely an object and a family of arrows, together with
    the function that computes the unique arrow from any other cone:
    
        Limit(object, arrows, unique_arrow_func, category)
        
    To obtain the limit of a diagram, one must create an instance of diagram
    and invoke the limit method on that instance. This method must return a
    Limit instance. 
    
    New: for ease of use, I added a functional interface to compute limits.
"""

from types import FunctionType, MethodType
from libutilitaspy.metaclassconfig import MetaClass
from libutilitaspy.data_structures.maps import Map
import diagrams


def limit(diagram):
    return diagram.limit()

def product(pair):
    assert isinstance(pair, diagrams.Pair)
    return pair.limit()

def equalizer(parallel_arrows):
    assert isinstance(parallel_arrows, diagrams.ParallelArrows)
    return parallel_arrows.limit()

def pullback(cospan):
    assert isinstance(cospan, diagrams.CoSpan)
    return cospan.limit()


class Cone(object):
    
    __metaclass__ = MetaClass

    def __init__(self, obj, arrows):
        assert all(obj == phi.source for phi in arrows)
        self.category = obj.category
        self.vertex = obj
        self.arrows = arrows
        self.arrows_by_target = Map((phi.target, phi) for phi in arrows)
    
    def valid(self, diagram):
        assert isinstance(diagram, diagrams.Diagram)
        return all(phi.target in diagram.objects for phi in self.arrows) \
            and all(diagrams.Triangle(phi, f, self.arrows_by_target[f.target]).commutes() \
                    for phi in self.arrows \
                    for f in phi.target.outgoing \
                    if f in diagram.arrows)
            

class Limit(Cone):
    
    __metaclass__ = MetaClass

    def __init__(self, obj, arrows, unique_arrow_func):
        """
        Parameters:
            obj: categories.Object
            arrows: Arrow list
            unique_arrow_func: Cone -> Arrow
        Preconditions:
            1) all(obj == phi.source for phi in arrows)
                # obj is the source of all arrows
            2) all(phi.target in diagram.objects for phi in arrows)
                # the target of all arrows are in the diagram
            3) Cone(obj, arrows, diagram).valid()
                # (obj, arrows) form a valid Cone
            4) 
                # unique_arrow_func is a function that given any cone (X,psi),
                # returns the unique arrow from X to obj that makes the
                # resulting diagrams commute
        """
        Cone.__init__(self, obj, arrows)
        assert type(unique_arrow_func) in (FunctionType, MethodType)
        self.unique_arrow_func = unique_arrow_func
    
    def unique_arrow(self, cone):
        assert isinstance(cone, Cone)
        u = self.unique_arrow_func(cone)
        assert u.source == cone.vertex
        assert u.target == self.vertex
        # We comment out the check for commutativity as checking composition
        # where arrows are or have functions is intractable
#        assert all(diagrams.Triangle(u, self.arrows_by_target[psi.target], psi).commutes() for psi in cone.arrows)
        return u


class Final(Limit):
    
    __metaclass__ = MetaClass
    
    def __init__(self, obj, unique_arrow_func):
        Limit.__init__(self, obj, [], unique_arrow_func)
    

class BinProduct(Limit):
    
    __metaclass__ = MetaClass

    def __init__(self, P, p1, p2, unique_arrow_func):
        Limit.__init__(self, P, [p1, p2], unique_arrow_func)
        self.proj1 = p1
        self.proj2 = p2

    
class Equalizer(Limit):
             
    __metaclass__ = MetaClass
    
    def __init__(self, Q, q, unique_arrow_func):
        Limit.__init__(self, Q, [q], unique_arrow_func)
        self.eq = q


class Pullback(Limit):
    
    __metaclass__ = MetaClass
    
    def __init__(self, Q, q1, q2, unique_arrow_func):
        Limit.__init__(self, Q, [q1, q2], unique_arrow_func)


