"""
This module implements the base classes of the *observer pattern* [GoF94]_.

.. [GoF94] E. Gamma, R. Helm, R. Johnson and J. Vlissides. Design Patterns: Elements of reusable Object-Oriented Software. 1994 
"""



class Observer:
    
    """
    An *observer* (also called *listener*) is any object which can be registered
    with some :py:class:`Observable` object to be notified whenever the
    Observable object is updated.
    """
    
    def notify(self, *args, **kwargs):
        """Method executed whenever any Observable where self is registered is updated.
        It should be overriden by a subclass."""
        pass


class Observable:
    
    """
    An *observable* object is any object that has a list of registered
    :py:class:`Observer` objects which are notified whenever the observable object is
    updated (a relevant change of state occurs).
    """
    
    def __init__(self):
        self.observers = []
    
    def register(self, observer):
        """
        Registers the observer object.
        
        :param Observer observer: The object to register.
        :post: after(self.observers) == before(self.observers) + [observer]
        """
        self.observers.append(observer)
    
    def deregister(self, observer):
        """
        Unregisters the observer object.
        
        :param Observer observer: The object to register.
        :pre: observer in self.observers
        :post: after(self.observers) == before(self.observers) - [observer]
        """
        self.observers.remove(observer)
    
    def update(self, *args, **kwargs):
        """
        Notifies all registered observers.
        """
        for observer in self.observers:
            observer.notify(*args, **kwargs)


